from collections import Iterable


__version__ = '0.2.1'


try:
    basestring
except NameError:
    basestring = str


class PermissionError(Exception):
    """Raised when permission is invalid."""


class RuleError(Exception):
    """Raised when there is no authorization rule for class of given object."""


class AuthorizationError(Exception):
    """Raised when user has no permission."""


class _Nothing(object):
    """Default class to associate authorization rules with."""


_nothing = _Nothing()  # default object to check permission for


class Authorization(object):
    """Container of authorization rules and checker of permissions."""
    def __init__(self):
        self._rules = {}

    def add_rule(self, rule_class, target_class=_Nothing):
        """Adds an authorization rule.

        :param rule_class: a class of authorization rule.
        :param target_class: (optional) a class
            or an iterable with classes to associate the rule with.
        """
        if isinstance(target_class, Iterable):
            for cls in target_class:
                self._rules[cls] = rule_class
        else:
            self._rules[target_class] = rule_class

    def rule_for(self, target_class):
        """Decorates and adds an authorization rule
        for a specified class(es) of objects.

        :param target_class: a class or an iterable with classes
            to associate the rule with.
        """
        def decorator(rule_class):
            self.add_rule(rule_class, target_class)
            return rule_class
        return decorator

    def rule(self, rule_class):
        """Decorates and adds an authorization rule."""
        self.add_rule(rule_class)
        return rule_class

    def check(self, user, permission, obj=_nothing):
        """Raises AuthorizationError when a user has no permission.

        :param user: a user.
        :param permission: permission to check.
        :param obj: (optional) an object to check permission for.
        """
        if not self.allows(user, permission, obj):
            raise AuthorizationError(
                'Can\'t {} object of class {}'.format(
                    permission, type(obj)))

    def allows(self, user, permission, obj=_nothing):
        """Checks that a user has permission. Returns True or False.

        :param user: a user.
        :param permission: permission to check.
        :param obj: (optional) an object to check permission for.
        """
        rule = self._get_rule(obj)

        if not isinstance(permission, basestring):
            return all(
                self._use_rule(rule, user, perm, obj)
                for perm in permission
            )

        return self._use_rule(rule, user, permission, obj)

    def _get_rule(self, obj):
        try:
            rule_class = self._rules[type(obj)]
        except KeyError:
            raise RuleError(
                'There is no authorization rule '
                'for class {}'.format(type(obj))
            )

        return rule_class()

    def _use_rule(self, rule, user, permission, obj):
        try:
            checker = self._get_checker(rule, permission)
        except AttributeError:
            raise PermissionError(
                'Unknown permission "{}" for object '
                'of class {}'.format(permission, type(obj))
            )

        if obj is _nothing:
            return checker(user)

        return checker(user, obj)

    @staticmethod
    def _get_checker(rule, permission):
        return getattr(rule, 'can_{}'.format(permission))

    def get_permissions(self, user, obj=_nothing):
        """Returns permissions of a user.

        :param user: a user.
        :param obj: (optional) an object to get permissions for.
        """
        rule = self._get_rule(obj)

        all_permissions = (
            attr[len('can_'):] for attr in dir(rule)
            if attr.startswith('can_')
        )

        return set(
            permission for permission in all_permissions
            if self.allows(user, permission, obj)
        )



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name='authorizeme',
    version=__import__('authorizeme').__version__,
    url='https://github.com/antonev/authorizeme',
    license='MIT',
    description='Authorization library for Python',
    author='Anton Evdokimov',
    author_email='antonevv@gmail.com',
    long_description='Authorization library for Python that can be used '
                     'with or without any Python web framework',
    py_modules=['authorizeme'],
    install_requires=[],
    zip_safe=False,
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
)



import pytest


class Blog(object):
    def __init__(self, writers=None):
        self.writers = writers or []


class User(object):
    def __init__(self, is_admin=False):
        self.is_admin = is_admin


class Article(object):
    def __init__(self, author):
        self.author = author


class BookReview(Article):
    pass


class MovieReview(Article):
    pass


@pytest.fixture(params=['declarative', 'imperative'])
def authorization(request):
    configuration_type = request.param

    from authorizeme import Authorization

    authz = Authorization()

    class AppRule(object):
        def can_add_blog(self, user):
            return user.is_admin

    class BlogRule(object):
        def can_add_article(self, user, obj):
            return user in obj.writers

        def can_add_author(self, user, obj):
            return user.is_admin

    class ArticleRule(object):
        def can_read(self, user, obj):
            return True

        def can_change(self, user, obj):
            return obj.author is user

        def can_rate(self, user, obj):
            return obj.author is not user

    if configuration_type == 'declarative':
        authz.rule(AppRule)
        authz.rule_for(Article)(ArticleRule)
        authz.rule_for([BookReview, MovieReview])(ArticleRule)
        authz.rule_for(Blog)(BlogRule)
    else:
        authz.add_rule(AppRule)
        authz.add_rule(ArticleRule, Article)
        authz.add_rule(ArticleRule, [BookReview, MovieReview])
        authz.add_rule(BlogRule, Blog)

    return authz


@pytest.fixture(params=[Article, BookReview, MovieReview])
def objects(request):
    article_class = request.param
    writer1, writer2 = User(), User()

    return {
        'blog': Blog(writers=[writer1, writer2]),
        'reader': User(),
        'writer1': writer1,
        'writer1_article': article_class(author=writer1),
        'writer2': writer2,
        'writer2_article': article_class(author=writer2),
    }


combinations = [
    # user, target, permission, result
    ('reader', 'blog', 'add_article', False),
    ('writer1', 'blog', 'add_article', True),
    ('writer1', 'writer1_article', 'read', True),
    ('writer1', 'writer1_article', 'change', True),
    ('writer1', 'writer1_article', 'rate', False),
    ('writer1', 'writer1_article', ['read', 'change'], True),
    ('writer1', 'writer1_article', ['read', 'change', 'rate'], False),
    ('writer1', 'writer2_article', 'read', True),
    ('writer1', 'writer2_article', 'change', False),
    ('writer1', 'writer2_article', 'rate', True),
    ('writer2', 'writer1_article', 'read', True),
    ('writer2', 'writer1_article', 'change', False),
    ('writer2', 'writer1_article', 'rate', True),
    ('writer2', 'writer2_article', 'read', True),
    ('writer2', 'writer2_article', 'change', True),
    ('writer2', 'writer2_article', 'rate', False),
    ('writer2', 'writer2_article', ['read', 'change'], True),
    ('writer2', 'writer2_article', ['read', 'change', 'rate'], False),
]


@pytest.mark.parametrize(
    'user_key, target_key, permission, result', combinations)
def test_allows(
        authorization, objects, user_key, target_key, permission, result):
    user = objects[user_key]
    target = objects[target_key]
    assert authorization.allows(user, permission, target) is result


@pytest.mark.parametrize(
    'user_key, target_key, permission, result', combinations)
def test_check(
        authorization, objects, user_key, target_key, permission, result):
    from authorizeme import AuthorizationError

    user = objects[user_key]
    target = objects[target_key]

    try:
        authorization.check(user, permission, target)
    except AuthorizationError:
        error_raised = True
    else:
        error_raised = False

    assert result is not error_raised


def test_rule_without_associated_class(authorization):
    admin = User(is_admin=True)
    assert authorization.allows(admin, 'add_blog')


@pytest.mark.parametrize('user_key, article_key, permissions', [
    ('writer1', 'writer1_article', {'read', 'change'}),
    ('writer1', 'writer2_article', {'read', 'rate'}),
    ('writer2', 'writer1_article', {'read', 'rate'}),
    ('writer2', 'writer2_article', {'read', 'change'}),
])
def test_get_permissions(
        authorization, objects, user_key, article_key, permissions):
    user = objects[user_key]
    article = objects[article_key]
    assert authorization.get_permissions(user, article) == permissions


def test_get_permissions_without_obj(authorization):
    admin = User(is_admin=True)
    assert authorization.get_permissions(admin) == {'add_blog'}


@pytest.fixture
def article():
    return Article(author=User())


def test_unknown_permission(authorization, article):
    from authorizeme import PermissionError

    with pytest.raises(PermissionError):
        authorization.allows(article.author, 'unknown_permission', article)


def test_unknown_permission_check(authorization, article):
    from authorizeme import PermissionError

    with pytest.raises(PermissionError):
        authorization.check(article.author, 'unknown_permission', article)


def test_unknown_class(article):
    from authorizeme import Authorization, RuleError

    authorization = Authorization()

    with pytest.raises(RuleError):
        authorization.allows(article.author, 'edit', article)


def test_unknown_class_check(article):
    from authorizeme import Authorization, RuleError

    authorization = Authorization()

    with pytest.raises(RuleError):
        authorization.allows(article.author, 'edit', article)
